#include <arpa/inet.h>
#include <iostream>
#include <netinet/in.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <ostream>
#include <sys/socket.h>
#include <unistd.h>

using std::cout, std::cin, std::endl, std::string;

/*
 * Remember to build with the flags "-L/usr/lib -lssl -lcrypto"
 */

int main() {

    struct Message {
        string content;
        string username;

        string toString() {
            return "User '" + username + "' wrote:\n" + content;
        }
    };

    // Initialize OpenSSL
    SSL_library_init();
    SSL_load_error_strings();

    // Create new SSL context
    SSL_CTX *ctx = SSL_CTX_new(TLS_client_method());
    if (!ctx) {
        cout << "Creation of SSL context failed" << endl;
    }

    // Load the server's certificate into context
    if (SSL_CTX_load_verify_locations(
            ctx, "./ssl/server.crt", nullptr) <= 0) {
            cout << "SSL load failed" << endl;
    }

    // Create SSL object
    SSL *ssl = SSL_new(ctx);
    if (!ssl) {
        cout << "Failed to create SSL object" << endl;
    }

    int clientSocket = socket(AF_INET, SOCK_STREAM, 0);
    SSL_set_fd(ssl, clientSocket);

    sockaddr_in serverAdress;
    serverAdress.sin_family = AF_INET;
    serverAdress.sin_port = htons(8080);
    serverAdress.sin_addr.s_addr = inet_addr("127.0.0.50");

    int connectStatus = connect(clientSocket, (struct sockaddr *)&serverAdress,
                                sizeof(serverAdress));
    if (connectStatus != 0) {
        cout << "Failed to connect";
        return connectStatus;
    } else {
        cout << "CONNECTED!" << endl;
    }

    // Perform SSL handshake
    if (SSL_connect(ssl) != 1) {
        cout << "SSL Handshake error" << endl;
    }

    while (true) {
        Message message;
        cout << "Type message to server: ";
        cin >> message.content;

        if (message.content == "exit") {
            cout << "BYE!" << endl;
            break;
        }

        cout << endl;
        message.username = "Client 1";

        SSL_write(ssl, message.toString().data(), strlen(message.toString().data()));
    }

    // Clean up
    SSL_free(ssl);
    SSL_CTX_free(ctx);
    ERR_free_strings();
    close(clientSocket);

    return 0;
}
